ShinyEnv:Use("ShinyStats")

local L = ShinyLocale

ShinyStats = LibStub("AceAddon-3.0"):NewAddon("ShinyStats", "AceConsole-3.0")
_G["ShinyStats"] = ShinyStats


function ShinyStats:OnInitialize()
	self.playerClass = nil
	self.playerSpecID = nil
	self.playerSpecIndex = nil
	self.tainted = false
	
	self.usefulStats = {}
	self.primaryStat = nil

	self.colors = nil
	self.patterns = nil

	self.tooltipsHooked = false
	self.eventsHooked = false

	ShinyDB:OnInitialize()
	ShinyCache:OnInitialize()
	ShinyConfig:OnInitialize()

	self:CreatePatterns()
end


function ShinyStats:OnEnable()
	self:UpdateClass()
	self:UpdateSpec()

	self:CreateColors()
	
	self:RegisterChatCommand("shiny", "ChatCommand")

	self:HookEvents()
	self:HookTooltips()

	self:UpdateIfTainted()
end

	
function ShinyStats:CreateColors()
	local colors = self.colors or {}
	local color

	for k,data in pairs(ShinyDB.db.profile.colors) do
		color = Color:CreateFromHEXA(data.value)
		color:SetEnabled(data.enabled)

		colors[k] = color
	end

	colors[SHINY_CLASS] = Color:CreateFromAHEX(RAID_CLASS_COLORS[self.playerClass].colorStr)

	self.colors = colors
end

	
function ShinyStats:CreatePatterns()
	local patterns = {}

	for k,v in pairs(PATTERN_CATEGORY_MAP) do
		patterns[k] = {}
	end

	for _, pattern in pairs(PATTERN_MAP) do
		for categoryName, categoryBit in pairs(PATTERN_CATEGORY_MAP) do
			if (pattern:HasCategory(categoryBit)) then
				table.insert(patterns[categoryName], pattern)
			end
		end
	end

	self.patterns = patterns
end


function ShinyStats:ReloadConfig()
	self.tainted = true

	self:CreateColors()
	self:UpdateIfTainted()
end


function ShinyStats:UpdateIfTainted()
	if (not self.tainted) then return end

	self:UpdateUsefulStats()
	self:UpdateDynamicColors()

	ShinyCache:Clear()
	self.tainted = false
end


function ShinyStats:UpdateClass()
	local _, class = UnitClass("player")

	self.playerClass = class
end


function ShinyStats:UpdateSpec()
	-- --[===[@debug@
	-- print("ShinyStats", "Updating Spec!")
	-- --@end-debug@]===]

	local specIndex = GetSpecialization()
	local lastPrimaryStat = self.primaryStat
	local lastSpecID = self.playerSpecID

	if (specIndex) then
		self.playerSpecIndex = specIndex
		self.playerSpecID = CLASS_SPEC_ID_MAP[self.playerClass][specIndex]
		self.primaryStat = select(7, GetSpecializationInfo(specIndex))
		-- local primaryStat = select(7, GetSpecializationInfo(spec, nil, nil, nil, UnitSex("player")))
		-- GetSpecializationInfo(GetSpecialization());
	end

	if (not (specIndex 
		and lastPrimaryStat == self.primaryStat
		and lastSpecID == self.playerSpecID)) then

		self.tainted = true
	end
end


function ShinyStats:UpdateUsefulStats()
	local list = self.usefulStats

	for k,v in pairs(STAT_CLASS_MAP) do
		list[k] = self:IsStatUseful(k, self.playerClass, self.playerSpecIndex) or false
	end
end


function ShinyStats:IsStatUseful(stat, class, spec)
	local stat = STAT_CLASS_MAP[stat][class]

	if (spec and type(stat) == "table") then
		return stat[spec]
	else
		return stat
	end
end


function ShinyStats:UpdateDynamicColors()
	self.colors[SHINY_PRIMARYSTAT] = self.colors[PRIMARY_STAT_MAP[self.primaryStat]]
end


function ShinyStats:ChatCommand(userInput)
	local userInput = userInput:lower()

	if (userInput == "dbg" or userInput == "debug") then
		ShinyDB.db.profile.options.DEBUG = not ShinyDB.db.profile.options.DEBUG
	elseif (userInput == "clearcache" or userInput == "clear") then
		ShinyCache:Clear()
	elseif (userInput == "import") then
		ShinyConfig:ImportConfig()
		ShinyCache:SetTainted(true)
		ShinyStats:CreateColors()
	else
		ShinyConfig:Show()
	end
end


function ShinyStats:IsRecipe(itemLink)
	return (itemLink:match(L["Recipe"]) 
		or itemLink:match(L["Pattern"]) 
		or itemLink:match(L["Plans"]) 
		or itemLink:match(L["Formula"]) 
		or itemLink:match(L["Technique"]) 
		or itemLink:match(L["Design"]) 
		or itemLink:match(L["Manual"]))
end


function ShinyStats:Escape(text)
	return text:gsub("|", "||")
end


function ShinyStats:Strip(text)
	text = text:gsub("|c%x%x%x%x%x%x%x%x", "")
	text = text:gsub("|r", "")

	return text
end


function ShinyStats:Shorten(text)
	local separator = ShinyDB:GetOption("SHORTEN_SEPARATOR")

	return text:gsub("|r" .. L["%sand%s"] .. "|c", "|r" .. separator .. " |c")
end


function ShinyStats:SplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	return itemLink:match("item:(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*)")
end


function ShinyStats:ValidateSplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	local result = {}

	for i,v in ipairs{self:SplitItemLink(itemLink)} do
		if (v and v ~= "") then
			result[i] = v
		end
	end
	
	return unpack(result)
end


function ShinyStats:ShouldSkip(region)
	local text = region:GetText()
	local r,g,b,a = region:GetTextColor()

	return (text:match(L["^Socket Bonus"]) and r < 0.51 and g < 0.51)				-- Inactive Socket Bonus
		or (text:match(L["^%(*%d*%)*%s*Set"]) and r < 0.51 and g < 0.51)			-- Inactive Set Bonus
		or text:match(L["^Legacy Set"])												-- Legacy Set
		or text:match(L["^<Shift"]) 												-- Shift-Click to ...
		or text:match(L["^<Made"]) 													-- Made by ...
		or text:match(L["^<Right"]) 												-- Right-Click to ...
		or text:match(L["^Season"]) 												-- Season (PvP)
		or text:match(L["^Requires"]) 												-- Requires ...
		or text:match(L["^Durability"]) 											-- Durability
		or text:match(L["^Reforged"])												-- Deprecated (Removed)
end


function ShinyStats:ParseRelic(text)
	local found, newColor = true

	if (CLASS_RELIC_MAP[self.playerClass][self.playerSpecIndex][text]) then
		newColor = SHINY_ARTIFACT_RELIC_ACTIVE

	elseif (CLASS_RELIC_MAP[self.playerClass][text]) then
		newColor = SHINY_ARTIFACT_RELIC_USABLE

	else
		newColor = SHINY_ARTIFACT_RELIC_UNUSABLE
	end

	return found, newColor
end


function ShinyStats:ParseEquip(text)
	local found, newText, newColor = false
	local effectTitle = self.colors[SHINY_EQUIP_PROC_TITLE]

	for _,data in ipairs(self.patterns[PATTERN_EQUIP]) do
		local pattern, colorKey = data:Unpack()
		local _, _, effect = string.find(text, pattern)

		if (effect) then
			newText = "|c" .. effectTitle:ToAHEX() .. PROC_PATTERN_REPLACEMENT_MAP[pattern] .. "|r\n" ..
					  effect:gsub("^%l", string.upper)

			-- if (ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES") and ShinyDB:GetOption("ENABLE_EQUIP")) then
			-- 	_, newText = self:ParseText(newText, PATTERN_PARTIAL, PATTERN_SPECIAL)
			-- end

			found = true
			newColor = colorKey
			break
		end
	end

	return found, newText, newColor
end


function ShinyStats:ParseLine(text, category)
	local found, newColor = false

	for _,data in ipairs(self.patterns[category]) do
		local pattern, colorKey = data:Unpack()

		-- print("ParseLine ("..category.."):", pattern, text)

		if (text:match(pattern)) then
			found = true
			newColor = colorKey

			-- print("\t", "Match found!")
			break
		end
	end

	return found, newColor
end


function ShinyStats:ParseText(text, ...)		-- categories
	local found, original, newText = false, text
	local replacements = {}

	local i = 0

	for _,category in ipairs{...} do
		for _,data in ipairs(self.patterns[category]) do
			local pattern, colorKey = data:Unpack()
			local color = self.colors[colorKey]
			local colorStr

			-- print("ParseText ("..category.."):", pattern, text)

			if (color and color:IsEnabled()) then
				text = text:gsub(pattern, 
					function(match)
						i = i + 1

						-- print("\t", "Match found! ("..i..")")
						replacements[i] = ("|c%s%s|r"):format(color:ToAHEX(), match)

						return "{"..i.."}"
					end)
			end
		end
	end

	if (#replacements > 0) then
		found = true
		newText = text:gsub(PARSE_TEXT_TEMPLATE_PATTERN, function(match)
				return replacements[tonumber(match)]
			end)
	end

	return found, newText or text
end


local lastItem, seen
function ShinyStats:ParseRegion(region, itemLink)
	if (not (region and region:GetObjectType() == "FontString")) then
		return false
	end

	local original = region:GetText()
	local text = original

	if (not text) then
		return false
	end

	local cacheResult = ShinyCache:IsEnabled() and ShinyCache:Get(original) or nil

	if (cacheResult == false) then
		-- print("String cached (ignored)")
		return false

	elseif (cacheResult) then
		-- print("String cached (parsed)")
		
		if (cacheResult.color and cacheResult.color:IsEnabled()) then
			-- print("\tRegion default text color set from cache")
			region:SetTextColor(cacheResult.color:ToDECA())
		end

		if (cacheResult.text) then
			-- print("\tRegion text set from cache")
			region:SetText(cacheResult.text)
		end

		return true
	end

	if (self:ShouldSkip(region)) then
		-- print("Region skipped, matching skip patterns")
		ShinyCache:Set(original, false)
		return false
	end

	local found, newText, newColor = false
	local skip = false

	if (text:match(L["^Enchant"])) then
		found, newColor = self:ParseLine(text, PATTERN_ENCHANT)

		if (ShinyDB:GetOption("SHORTEN_ENCHANT")) then
			newText = self:Shorten(newText or text)
		end
		
	elseif (text:match(L["^Equip"])) then
		if (ShinyDB:GetOption("ENABLE_EQUIP_TRANSFORMATION")) then
			found, newText, newColor = self:ParseEquip(text)
		end

		if (ShinyDB:GetOption("SHORTEN_EQUIP")) then
			newText = self:Shorten(newText or text)
		end

		if (ShinyDB:GetOption("ENABLE_EQUIP") and ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
			partialFind, newText = self:ParseText(newText or text, PATTERN_PARTIAL, PATTERN_SPECIAL)
		end

		found = found or partialFind
		skip = true

	elseif (text:match(L["^Use"])) then
		found, newColor = self:ParseLine(text, PATTERN_USE)

		if (ShinyDB:GetOption("SHORTEN_USE")) then
			newText = self:Shorten(newText or text)
		end

	elseif (text:match(L["^Chance"])) then
		found, newColor = self:ParseLine(text, PATTERN_CHANCE)

		if (ShinyDB:GetOption("SHORTEN_CHANCE")) then
			newText = self:Shorten(newText or text)
		end

	elseif (text:match(L["^%a+ Artifact Relic$"])) then
		found, newColor = self:ParseRelic(text)

	else
		found, newColor = self:ParseLine(text, PATTERN_LINE)

		if (ShinyDB:GetOption("SHORTEN_ENCHANT")) then
			newText = self:Shorten(newText or text)
		end

		if (ShinyDB:GetOption("ENABLE_STAT_STRIP_COLORS")) then
			newText = self:Strip(newText or text)
		end
	end

	if (not found and not skip and ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
		found, newText = self:ParseText(text, PATTERN_PARTIAL, PATTERN_SPECIAL)

		if (ShinyDB:GetOption("SHORTEN_SOCKET_GEM")) then
			newText = self:Shorten(newText or text)
		end
	end

	-- print("Found:", found)

	if (found) then
		local color

		-- print("Found", newColor)
		if (ShinyDB:GetOption("ENABLE_STAT_DIM_USELESS") and newColor and STAT_CLASS_MAP[newColor] and not self.usefulStats[newColor]) then
			color = self.colors[SHINY_USELESS]
		else
			color = newColor and self.colors[newColor]
		end

		if (color and color:IsEnabled()) then
			region:SetTextColor(color:ToDECA())
		end

		if (newText ~= original and newText) then
			region:SetText(newText)
		end

		ShinyCache:Set(original, {
				color = color,
				text = newText ~= original and newText
			})

	else
		ShinyCache:Set(original, false)
	end

	-- if (cacheData) then
	-- 	print("CacheData:", cacheData.color, cacheData.text)
	-- end

	if (ShinyDB:GetOption("DEBUG")) then
		if (lastItem ~= itemLink) then
			seen = setmetatable({}, { __mode = "kv" })
			lastItem = itemLink
			print(itemLink)
			seen[region] = true
		elseif (not seen[region]) then
			local pText = newText ~= original and newText and "-->    "..newText or newColor and newColor.id or ""
			print(string.format("%-15s%-45s    %s    %s    %s",found and "(found)" or "", text, pText, ShinyStats:Escape(text), ShinyStats:Escape(pText)))
			seen[region] = true
		end
	end

	return found
end


function ShinyStats:ExtendTooltip(tip, itemLink)
	self:UpdateIfTainted()

	if (ShinyCache:IsTainted()) then
		ShinyCache:Clear()
	end

	for _,region in ipairs{tip:GetRegions()} do
			self:ParseRegion(region, itemLink)
	end

	tip:Show()
end


function ShinyStats:HookEvents()
  if (self.eventsHooked) then return end
	self.eventsHooked = true

	self.listener:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self.listener:RegisterEvent("PLAYER_TALENT_UPDATE")
end


function ShinyStats:HookTooltips()
  if (self.tooltipsHooked) then return end
  self.tooltipsHooked = true

  GameTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)

  ItemRefShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)

  ShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)

  if (AtlasLoot and AtlasLoot.db.Tooltip.tooltip == "AtlasLootTooltip" and AtlasLootTooltip) then
  	AtlasLootTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  end


  if LinkWrangler and LinkWrangler.RegisterCallback then
    LinkWrangler.RegisterCallback("ShinyStats", function(frame,link)
      ShinyStats:ExtendTooltip(frame,link)
    end, "item")
  end

end


function ShinyStats.TooltipHook(tip)
	local _, itemLink = tip:GetItem()

	if (itemLink) then
		return ShinyStats:ExtendTooltip(tip, itemLink)
	end
end
